"
 COPYRIGHT (c) 1997 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"


View subclass:#TabView
	instanceVariableNames:'leftInset list listHolder selection action tabStyle
		sizeChangedNotify useIndex selectionToFirstLine maxLineNr'
	classVariableNames:''
	poolDictionaries:''
	category:'Views-Interactors'
!

!TabView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

!

documentation
"
    implements the tabs-view component of a noteBook.
    May also be used on its own (without a surrounding noteBook).

    The functionality is basically the same as provided by a
    PopUpList or SelectionInListView, in that a valueHolder
    gets a value assigned corresponding to the selected tab
    from a list of possible tabs.

    [author:]
        Claus Atzkern

    [see also:]
        NoteBookView
        SelectionInListView PopUpList ValueHolder
"

!

examples
"
                                                                                [exBegin]                                      
    |top sel view|

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:300@200.


    sel  := self origin:0.0 @ 0.0 corner:1.0 @ 0.7 in:top.
    view := View origin:0.0 @ 0.7 corner:1.0 @ 1.0 in:top.
    view viewBackground:(sel styleAt:#selectedColor).

    sel list:#( 'Button'
                'Toggle'
                'Panel'
                'Text'
                'View'
                'Combo View'
                'Menu'
                'Scroller'
              ).
    sel action:[:aName|
        Transcript showCR:aName
    ].
    top open.
                                                                                [exEnd]                                      



                                                                                [exBegin]                                      
    |top sel view y|

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:450@300.

    y    := 150.
    sel  := self origin:0.0 @ 0.0 corner:1.0 @ y in:top.
    sel horizontalInset:10.
    view := NoteBookFrameView origin:0.0 @ y  corner:1.0 @ 1.0 in:top.
    view horizontalInset:10.
    view bottomInset:10.
    view level:2.
    view viewBackground:(Image fromFile:'bitmaps/gifImages/garfield.gif').

    sel styleAt:#selectedColor    put:(view viewBackground).
    sel styleAt:#unselectedColor  put:(Color grey:60).
    sel styleAt:#hspace           put:10.
    sel styleAt:#expandSelection  put:9@7.

    sel list:#( 'Button'
                'Toggle'
                'Claus Atzkern'
                'Panel'
                'Text'
                'Peter'
                'Loechgau'
                'Text'
              ).
    sel useIndex:true.
    sel action:[:aName|
        Transcript showCR:aName
    ].

    top open.
                                                                                [exEnd]                                      




                                                                                [exBegin]                                      
    |top sel view ctr l|

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:420@100.


    sel  := self origin:0.0 @ 0.0 corner:1.0 @ 1.0 in:top.

    l := #(     'SBrowser'
                'FBrowser'
                'Debugger'
              ).
    sel list:(l collect:[:n | Image fromFile:'bitmaps/' , n , '.xbm']).

    ctr := 0.
    sel action:[:indexOrNil|
        Transcript showCR:indexOrNil
    ].
    top open.
                                                                                [exEnd]                                      

                                                                                [exBegin]                                      
    |top sel view l top2 s top3 p|

    l := SelectionInList new.
    l list:#('foo' 'bar' 'baz').
    l selectionIndex:1.

    top2 := StandardSystemView new.
    top2 extent:100@200.
    s := SelectionInListView origin:0.0@0.0 corner:1.0@1.0 in:top2.
    s model:l.
    top2 open.

    top3 := StandardSystemView new.
    top3 extent:100@200.
    s := PopUpList in:top3.
    s model:l.
    top3 open.

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:400@100.


    sel  := self origin:0.0 @ 0.0 corner:1.0 @ 1.0 in:top.
    sel model:(l selectionIndexHolder).
    sel listHolder:(l listHolder).

    sel useIndex:true.
    sel action:[:indexOrNil|
        Transcript showCR:indexOrNil
    ].
    top open.
                                                                                [exEnd]


                                                                                [exBegin]
    |top sel view y|

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:500@200.

    y    := 60.
    sel  := self origin:0.0 @ 0.0 corner:1.0 @ y in:top.
    sel horizontalInset:10.
    view := NoteBookFrameView origin:0.0 @ y  corner:1.0 @ 1.0 in:top.
    view horizontalInset:10.
    view bottomInset:10.
    view level:2.
    view viewBackground:(Image fromFile:'bitmaps/gifImages/garfield.gif').

    sel tabWidget:#Window.
    sel styleAt:#selectedColor   put:(view viewBackground).
    sel styleAt:#unselectedColor put:(Color grey:60).

    sel list:#( 'Button'
                'Toggle'
                'Claus'
                'Panel'
                'Text'
                'Peter'
                'Text'
              ).
    top open.
                                                                                [exEnd]
"
!

test
    |top sel view|

    top := StandardSystemView new
        label:'UI-Selection Panel';
        extent:300@200.


    sel  := self origin:0.0 @ 0.0 corner:1.0 @ 0.7 in:top.
    view := View origin:0.0 @ 0.7 corner:1.0 @ 1.0 in:top.
    view viewBackground:(sel styleAt:#selectedColor).

    sel list:#( 'Button'
                'Toggle'
                'Panel'
                'Text'
                'View'
                'Combo View'
                'Menu'
                'Scroller'
              ).
    sel action:[:aName|
        Transcript showCR:aName
    ].
    top open.

! !

!TabView class methodsFor:'defaults'!

defaultTabWidget
    ^ #Folder
! !

!TabView methodsFor:'accessing'!

action:oneArgBlock
    "set the action block to be performed on select; the argument to
     the block is the selected index or nil in case of no selection.
    "
    action := oneArgBlock.

!

list
    "return the list
    "
    ^ list
!

list:aList
    "set the list
    "
    |widget|

    selection := nil.
    list := nil.

    aList size ~~ 0 ifTrue:[
        widget := tabStyle at:#widget.
        list   := aList collect:[:aLabel|widget label:aLabel for:self].
    ].

    self shown ifTrue:[
        self recomputeList.
        self redrawFull.
    ]
!

listHolder
    "returns the list holder
    "
    ^ listHolder
!

listHolder:aValueHolder
    "change the list holder
    "
    listHolder notNil ifTrue:[
        listHolder removeDependent:self. 
    ].
    listHolder := aValueHolder.
    listHolder notNil ifTrue:[
        listHolder addDependent:self.
        self list:listHolder value.
        self selection:model value.
    ].
!

model:aValueHolder
    super model:aValueHolder.
    model notNil ifTrue:[
        self selection:(model value)
    ]
!

preferredExtent
    "compute max extent x/y
    "
    |x y ovl size hspace vspace|

    ovl    := tabStyle at:#leftOverLap.
    hspace := tabStyle at:#hspace.
    vspace := tabStyle at:#vspace.
    x := ovl + leftInset.
    y := 0.

    size := list size.
    size == 0 ifTrue:[ ^ 100@32 ].

    list do:[:aTab||ext|
        ext := aTab preferredExtent.
        ext y > y ifTrue:[
            y := ext y.
        ].
        x := x + ext x - ovl + hspace.
    ].
    y := y + (size * vspace) + (tabStyle at:#expandSelection) y.
  ^ x @ y.

!

preferredHeight
    "returns preferred height
    "
    list notNil ifTrue:[
      ^ (self extent y) - ((list first) origin y) + ((tabStyle at:#expandSelection) y)
    ].
    ^ 0
!

useIndex:aBoolean
    "set/clear the useIndex flag. If set, both actionBlock and change-messages
     are passed the index(indices) of the selection as argument. 
     If clear, the value(s) (i.e. the selected string) is passed.
     Default is false."

    useIndex := aBoolean


! !

!TabView methodsFor:'accessing style'!

styleAt:anIdentifier
    "returns value for an identifier
    "
    ^ tabStyle at:anIdentifier
!

styleAt:anIdentifier put:something
    "set value for an identifier
    "
    tabStyle at:anIdentifier put:something.

    list notNil ifTrue:[
        list first styleChanged:anIdentifier to:something
    ].
!

tabWidget
    "returns the tab widget class
    "
    |widget|

    widget := tabStyle at:#widget.
    widget := widget nameWithoutPrefix asSymbol.
  ^ widget
!

tabWidget:aWidget
    "change tab widget class
    "
    |widget|

    (tabStyle at:#widget) = aWidget ifFalse:[
        (widget := TabWidget widgetClass:aWidget) notNil ifTrue:[
            tabStyle := widget tabStyleOn:self.

            self shown ifTrue:[
                self recomputeList.
                self redrawFull.
            ]
        ]
    ].

! !

!TabView methodsFor:'change & update'!

update:something with:aParameter from:changedObject
    "any of my model changed
    "
    changedObject == model ifTrue:[
        self selection:model value
    ] ifFalse:[
        changedObject == listHolder ifTrue:[
            self list:(listHolder value)
        ]
    ]
! !

!TabView methodsFor:'drawing'!

paintColor:aColorSymbol
    "set paint to a color from the style identified by its symbol
    "
    self paint:(tabStyle at:aColorSymbol)
!

redrawFull
    "redraw list and boarders
    "
    |size x1 x2 y1 y2 space sel|

    size := list size.

    (size ~~ 0 and:[self shown]) ifFalse:[
        ^ self
    ].
    space := tabStyle at:#vspace.

    self paint:(self viewBackground).
    self clear.

 "/ redraw view boarders
    space ~~ 0 ifTrue:[
        x1 := tabStyle at:#hspace.
        y1 := self extent y - space.
        x2 := self extent x.

        self paintColor:#labelColor.

        (size+1) timesRepeat:[
            self displayLineFrom:(0 @ y1) to:(x2 @ y1).
            y1 := y1 - space.
        ]
    ].

 "/ redraw list
    maxLineNr to:1 by:-1 do:[:i|self redrawLineAt:i].
    selection notNil ifTrue:[
        sel := selection.
        selection := nil.
        self setSelection:sel.
    ].
!

redrawLineAt:aLineNr
    "redraw one line
    "
    list do:[:aTab|aTab lineNr == aLineNr ifTrue:[aTab redraw:false]]
!

redrawX:x y:y width:w height:h
    "a region must be redrawn
    "
    self redrawFull
! !

!TabView methodsFor:'event handling'!

buttonPress:button x:x y:y
    "a button is pressed
    "
    |lnNr index|

    list notNil ifTrue:[
        index := nil.
        lnNr  := 99999.

     "/ find tab under point; update selection

        list keysAndValuesDo:[:aKey :aTab|
            (aTab containsPoint:(x@y)) ifTrue:[
                aTab lineNr < lnNr ifTrue:[
                    index := aKey.
                    lnNr  := aTab lineNr.
                ]
            ]
        ].
        self selection:index
    ]
!

keyPress:aKey x:x y:y
    "selection might change; look for corresponding list entry
    "
    |size newSel|

    (size := list size) ~~ 0 ifTrue:[
        aKey == #CursorLeft ifTrue:[
            (selection isNil or:[selection == size]) ifTrue:[
                newSel := 1
            ] ifFalse:[
                newSel := (selection + 1)
            ]
        ] ifFalse:[
            aKey == #CursorRight ifTrue:[
                (selection isNil or:[selection == 1]) ifTrue:[
                    newSel := size
                ] ifFalse:[
                    newSel := (selection - 1)
                ]
            ]
        ].
        newSel notNil ifTrue:[
            self selection:newSel
        ]
        
    ].

!

sizeChanged:how
    "size of view changed 
    "
    list notNil ifTrue:[
        self recomputeList.
        sizeChangedNotify ifFalse:[
            sizeChangedNotify := true.
            self changed:#preferredHeight.
            sizeChangedNotify := false.
        ]
    ].

! !

!TabView methodsFor:'initialization'!

initialize
    "setup default attributes
    "
    |widget|

    super initialize.
    widget   := TabWidget widgetClass:(self class defaultTabWidget).
    tabStyle := widget tabStyleOn:self.
    leftInset := 0.
    sizeChangedNotify := false.
    useIndex := false.
    selectionToFirstLine := false.
! !

!TabView methodsFor:'layout'!

fitLineAt:aLnNr
    "fit line at a lineNr
    "
    |dltExtX dltOrgX tab sIdx eIdx width|

    eIdx := list findFirst:[:aTab| aTab lineNr == aLnNr].
    eIdx == 0 ifTrue:[^ false].

    sIdx  := list findLast:[:aTab| aTab lineNr == aLnNr].
    tab   := list at:eIdx.
    width := self extent x.

    dltExtX := width - ((tab origin x) + (tab extent x)).
    dltExtX := dltExtX // (sIdx - eIdx + 1).
    dltOrgX := 0.

    dltExtX > 0 ifTrue:[
        sIdx to:eIdx by:-1 do:[:i|
            tab := list at:i.
            tab extent x:((tab extent x) + dltExtX).
            tab origin x:((tab origin x) + dltOrgX).
            dltOrgX := dltOrgX + dltExtX.
        ]
    ].
    tab := list at:eIdx.
    tab extent x:(width - tab origin x).
  ^ true



!

recomputeFixedLayout
    "recompute tabs
    "
    |vspace tab lnNr deltaY vspCtr size x y hspace width ovlap noLn|

    size := list size.
    hspace := tabStyle at:#hspace.
    width  := self extent x - leftInset.
    vspace := tabStyle at:#vspace.
    vspCtr := vspace.

    deltaY := list first preferredExtent y.
    x := leftInset.
    y := self extent y.
    lnNr  := 1.
    ovlap := tabStyle at:#leftOverLap.
    maxLineNr := 1.

    list reverseDo:[:aTab||e n|
        e := aTab preferredExtent.
        n := e x + x - ovlap.

        n > width ifTrue:[
            maxLineNr := maxLineNr + 1.
            lnNr := lnNr + 1.
            x := leftInset.
            y := y - deltaY.
            n := e x + x - ovlap.
            vspCtr := 0.
        ].
        aTab lineNr:lnNr.
        aTab origin:x@(y - deltaY - vspCtr) extent:e inset:vspCtr.
        x := n + hspace.
        size := size - 1.
        vspCtr := vspCtr + vspace.
    ].

    "/ fit lines
    noLn := 1.
    [self fitLineAt:noLn] whileTrue:[noLn := noLn + 1].
!

recomputeList
    "recompute tabs
    "
    |numCols lnNr tabOffset size extent width x y 
     ovlap extX hspace origin ctr bottomInset vspace|

    size := list size.
    size == 0 ifTrue:[^ self].

    (tabStyle at:#fixedLabelSize) ifTrue:[
        ^ self recomputeFixedLayout
    ].
    extent := self preferredExtent.
    (extent x) <= (self extent x) ifTrue:[
        ^ self recomputeFixedLayout
    ].

 "/ compute preferred extent

    x     := 10.
    y     := 10.
    hspace:= tabStyle at:#hspace.
    ovlap := tabStyle at:#leftOverLap.

    list do:[:aTab|
        extent := aTab preferredExtent.
        extent x > x ifTrue:[x := extent x].
        extent y > y ifTrue:[y := extent y].
    ].
    width   := self extent x - ovlap - leftInset.
    extent  := x@y.
    extX    := x - ovlap.
    numCols := width // (extX + hspace).

 "/ calculate tabOffset, numCols and lnNr

    numCols < 2 ifTrue:[
        numCols := 1.
        maxLineNr := size.
    ] ifFalse:[
        numCols > size ifTrue:[
            numCols := size
        ].

        maxLineNr := (size + numCols - 1) // numCols.
        tabOffset := (maxLineNr * hspace) + extX.

        (numCols * tabOffset) > width ifTrue:[
            numCols := numCols - 1.
            maxLineNr := (size + numCols - 1) // numCols.
        ]
    ].
    tabOffset := width // numCols.

 "/ calculate extent x
    extent x:(tabOffset - (maxLineNr * hspace) + ovlap).

    vspace := tabStyle at:#vspace.
    bottomInset := vspace.
    y    := self extent y - vspace - extent y.
    x    := leftInset.
    ctr  := 1.
    lnNr := 1.

    list reverseDo:[:aTab|
        aTab lineNr:lnNr.
        aTab origin:x@y extent:extent inset:bottomInset.
        bottomInset := bottomInset + vspace.

        (ctr := ctr + 1) > numCols ifTrue:[
            origin := (list at:size) origin.
            size   := size - numCols.
            y      := origin y - extent y.
            x      := origin x + hspace.
            ctr    := 1.
            lnNr   := lnNr + 1.
        ] ifFalse:[
            x := x + tabOffset.
            y := y - vspace.
        ]
    ]
! !

!TabView methodsFor:'selection'!

lineToBottomWithNumber:aLineNr
    "exchange the first line with the received line number
    "
    |tabN tab1 orgY|

    tab1 := list at:(list findLast:[:aTab|aTab lineNr == 1]).
    tabN := list at:(list findLast:[:aTab|aTab lineNr == aLineNr]).
    orgY := tab1 origin y - tabN origin y.

    list do:[:aTab||ln|
        ln := aTab lineNr.
        ln == 1 ifTrue:[
            aTab origin y:(aTab origin y - orgY).
            aTab lineNr:aLineNr.
        ] ifFalse:[
            ln == aLineNr ifTrue:[
                aTab origin y:(aTab origin y + orgY).
                aTab lineNr:1.
            ]
        ]
    ]
!

selection
    "return the selection index or nil
    "
    useIndex ifTrue:[
        ^ selection
    ].

    (selection notNil and:[list size >= selection]) ifTrue:[
        ^ (list at:selection) label
    ].
    ^ nil
!

selection:anIndexOrNil
    "change the selection to index or nil. The model and/or actionBlock is notified
    "
    |sel|

    sel := selection.
    self setSelection:anIndexOrNil.

    (sel ~~ selection) ifTrue:[
        sel := self selection.

        model notNil ifTrue:[
            model value:sel
        ].

        action notNil ifTrue:[
            action value:sel
        ]
    ]

!

setSelection:anIndexOrNil
    "change the selection to index or nil. No notifications are raised
    "
    |index lnNr|

    list isNil ifTrue:[
        ^ self
    ].
    anIndexOrNil isString ifTrue:[
        "some model is using a string and not an index
        "
        index := list findFirst:[:aTab|aTab label = anIndexOrNil].
    ] ifFalse:[
        index := anIndexOrNil
    ].
        
    selection == index ifTrue:[
        ^ self
    ].
    self shown ifFalse:[
        selection := index.
      ^ self
    ].

    (     index notNil
     and:[(lnNr := (list at:index) lineNr) > 1
     and:[(tabStyle at:#selectionAtBottom)
     and:[(tabStyle at:#vspace) == 0]]]
    ) ifTrue:[
        self lineToBottomWithNumber:lnNr.
        selection := 1.
    ].

    selection notNil ifTrue:[
        selection := nil.
        self redrawFull
    ].
    selection := index.
    self showSelected.
!

showSelected
    "draw current selection as selected
    "
    |leftExtX rightExtX vspace tab size savOrg savExt expSel savIns setIns newExt x newOrg|

    size := list size.

    (selection notNil and:[self shown and:[size ~~ 0]]) ifFalse:[
        ^ self
    ].
    tab       := list at:selection.
    savOrg    := tab origin.
    savExt    := tab extent.
    savIns    := tab inset.
    expSel    := tabStyle at:#expandSelection.
    rightExtX := expSel x.

    (vspace := tabStyle at:#vspace) == 0 ifTrue:[
        leftExtX := rightExtX := rightExtX//2.
        setIns := 0
    ] ifFalse:[
        leftExtX := 0.
        setIns   := savIns
    ].

    newExt := savExt + ((leftExtX + rightExtX)@0).
    newOrg := savOrg - ( leftExtX @ (expSel y)).

    (x := newOrg x) < 0 ifTrue:[
        newExt x:(newExt x + x).
        newOrg x:0.
        x := 0.
    ].

    (x + newExt x) > (self extent x) ifTrue:[
        newExt x:((self extent x) - x)
    ].

    tab origin:newOrg extent:newExt inset:setIns.
    tab redraw:true.
    tab origin:savOrg extent:savExt inset:savIns.

    x := tab lineNr.
    [(x := x - 1) ~~ 0] whileTrue:[self redrawLineAt:x].
! !

!TabView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/TabView.st,v 1.5 1997-04-08 15:23:53 ca Exp $'
! !
